import click
from click.core import Group
from netlab.auth import load_config, add_config, remove_config

from .utils import client, output


@click.command(cls=Group, invoke_without_command=True)
@click.option('--system', default=None, help="Only show settings from a specific NETLAB+ system.")
@click.pass_context
def config(ctx, system):
    """ Manage config.json settings """
    if ctx.invoked_subcommand is None:
        systems = {}
        system_configs = load_config(system)

        if not system_configs:
            click.secho("No config settings present. See below on how to add them.", fg='red', bg='white')
            click.echo()
            click.echo(ctx.get_help())
            ctx.exit(0)

        if system:
            systems[system] = system_configs
        else:
            systems = system_configs

        for system, conf in systems.items():
            click.echo('System: {}'.format(click.style(system, fg='cyan')))
            output(conf)
            click.echo()
        click.echo("To learn more about modifying these settings, display the help text by running:")
        click.secho("    {} {}".format(ctx.command_path, ctx.help_option_names[0]), fg='cyan')


@config.command()
@click.option('--system', default='default', help="A friendly name to identify this NETLAB+ system. Leave blank for "
                                                  "'default'.")
@click.option('--host', prompt=True, help="IP Address or hostname of remote NETLAB+ API.")
@click.option('--user', prompt=True, help="User account ID from NETLAB+ system.")
@click.option('--token', prompt=True, help="Token obtained from NETLAB+ system.")
@click.option('--port', type=click.INT, help="TCP port to connect to the remote NETLAB+ API. Defaults to 9000.")
@click.option('--timeout', prompt=True, default=10, help="The amount of time in seconds of no response from the "
                                                         "socket before it will close. Defaults to 10.")
def add(system, host, user, token, port, timeout):
    """ Add a new system with all required settings """
    current_config = load_config()
    if system in current_config:
        click.confirm('\nYou already have a system named {}. If you continue, the existing one will be overwritten. '
                      'Are you sure?'.format(click.style(system, fg='cyan')), abort=True)

    new_config = {
        'host': host,
        'user': user,
        'token': token,
        'timeout': timeout
    }

    if port:
        new_config['port'] = port

    add_config(system, new_config)
    click.secho("Success!", fg='green')


@config.command()
@click.pass_context
@click.option('--system', help="The NETLAB+ system to remove. Required.")
@click.option('--confirm', is_flag=True, help="Skips the prompt to confirm removal. This will remove the "
                                              "config without warning. {warning}"
              .format(warning=click.style("Use at your own risk!", bg='white', fg='red')))
@click.option('--all', 'all_systems', is_flag=True, help="Removes all systems and settings. Resets your "
                                                         "config.json back to nothing.")
def remove(ctx, system, confirm, all_systems):
    """ Remove all settings for a system """
    if not system and not all_systems:
        click.echo(ctx.get_help())
        ctx.exit(0)

    if all_systems:
        system = None
        confirm_message = 'You are about to remove the settings for {} the systems. Are you sure?'.format(
            click.style("all", fg='red'))
        end_message = 'All systems have been removed.'
    else:
        confirm_message = 'You are about to remove the settings for the {} system. Are you sure?'.format(
            click.style(system, fg='cyan'))
        end_message = 'The {} system has been removed.'.format(click.style(system, fg='cyan'))

    if not confirm:
        click.secho("\nWARNING!", fg='red')
        click.confirm(confirm_message, abort=True)

    remove_config(system)
    click.echo(end_message)


@config.command(name='set')
@click.option('--system', default='default', help="The NETLAB+ system to modify the settings.")
@click.argument('settings', nargs=-1)
def set_value(system, settings):
    """ Set one or more settings for a specific system """
    current_config = load_config(system)

    new_settings = {}
    for setting in settings:
        key, value = setting.split('=')
        new_settings[key.lower()] = value

    new_config = {}
    new_config.update(current_config)
    new_config.update(new_settings)
    add_config(system, new_config)

    click.secho("Success!", fg='green')
    click.echo("Updated the config file with the following:\n")
    output(new_settings, headers=['Setting', 'New Value'])


@config.command()
@click.option('--system', default='default', help="The NETLAB+ system to get settings from.")
@click.option('--oneline', is_flag=True)
@click.argument('settings', nargs=-1)
def get(system, oneline, settings):
    """ Get the current value of one or more settings for a specific system """
    current_config = load_config(system)
    gathered_settings = []

    for setting in settings:
        try:
            current_setting_value = current_config[setting]
        except KeyError:
            # if the setting is not present in config.json, then ommit it from the results
            pass
        else:
            gathered_settings.append([setting, current_setting_value])

    if oneline:
        settings_string = " ".join(["=".join(setting) for setting in gathered_settings])
        click.echo(settings_string)
    else:
        click.echo('System: {}\n'.format(click.style(system, fg='cyan')))
        output(gathered_settings, headers=['Setting', 'Value'])


@config.command()
@click.option('--system', default='default', help="The NETLAB+ system to test connection to.")
def test(system):
    """ Test that the system settings can connect successfully to the server"""
    results = client(system).system_status_get()

    click.secho("Test connection successful!\n", fg='green')
    output(results)



import functools
from distutils.version import StrictVersion

from .errors.common import InvalidVersion
from .enums import HDRSeverity


def compare_version(v1, v2):
    s1 = StrictVersion(v1)
    s2 = StrictVersion(v2)
    if s1 == s2:
        return 0
    elif s1 > s2:
        return -1
    else:
        return 1


def version_lt(v1, v2):
    return compare_version(v1, v2) > 0


def version_gte(v1, v2):
    return not version_lt(v1, v2)


def minimum_version(version):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(self, *args, **kwargs):
            if version_lt(self._server_version, version):
                raise InvalidVersion(
                    '{0} is only available on NETLAB+ systems version {1} or later.'.format(
                        f.__name__, version
                    )
                )
            return f(self, *args, **kwargs)
        return wrapper
    return decorator


def hdr_result(results, severity_level=HDRSeverity.WARN):
    # TODO when card #1109 resolves, include hdr group id in response
    response = {'status': results.get('status_t'), 'value': results.get('value'), 'events': []}
    for event in results.get('events'):
        if event['severity'] <= severity_level.value:
            response['events'].append(event)
    return response



import logging
import socket
import ssl
import sys

from .errors.common import AuthenticationError, NetlabError
from .serializer import serialize, deserialize

log = logging.getLogger(__name__)


class Connection(object):
    def __init__(self, user, token, host, port, socket_timeout=None, retry_on_timeout=False):
        self.user = user
        self.token = token
        self.host = host
        self.port = int(port)
        self.socket_timeout = socket_timeout
        self.retry_on_timeout = retry_on_timeout
        self._sock = None
        self._file = None

    def __del__(self):
        try:
            self.disconnect()
        except Exception:
            pass

    def _create_socket(self):
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    def connect(self):
        if self._sock:
            return
        try:
            self._create_socket()
            self._sock.settimeout(self.socket_timeout)
            self._sock.connect((self.host, self.port))
        except socket.timeout as e:
            log.debug(e)
            raise
        except socket.error as e:
            log.debug(e)
            raise

        self._file = self._sock.makefile()

        try:
            self.on_connect()
        except NetlabError:
            # clean up after any error in on_connect
            self.disconnect()
            raise

    def on_connect(self):
        auth_request = {
            'id': 1,
            'jsonrpc': '2.0',
            'method': 'user.authenticate',
            'params': {'user': self.user, 'token': self.token}
        }
        self.send(serialize(auth_request))
        response = deserialize(self.read_response())

        if 'error' in response:
            raise AuthenticationError("Login failed")

    def disconnect(self):
        self.on_disconnect()

        if self._sock is None:
            return
        try:
            if self._file is not None:
                self._file.close()
                self._file = None
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
        except socket.error:
            pass
        self._sock = None

    def on_disconnect(self):
        pass

    def send(self, request):
        if not self._sock:
            self.connect()
        try:
            serialized_data = request.encode('utf-8')  # socket needs bytestring
            # newline indicates end of RPC request
            self._sock.sendall(serialized_data + b"\n")
        except socket.timeout:
            self.disconnect()
            raise TimeoutError("Timeout writing to socket")
        except socket.error:
            e = sys.exc_info()[1]
            self.disconnect()
            if len(e.args) == 1:
                errno, errmsg = 'UNKNOWN', e.args[0]
            else:
                errno = e.args[0]
                errmsg = e.args[1]
            raise ConnectionError("Error %s while writing to socket. %s." %
                                  (errno, errmsg))
        except:
            self.disconnect()
            raise

    def read_response(self):
        try:
            response = self._file.readline()
            if response == "":
                raise ConnectionError('The remote server closed the connection')
        except:
            self.disconnect()
            raise
        return response


class SSLConnection(Connection):
    def _create_socket(self):
        # overriding this method to do SSL specific setup of the socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock = ssl.wrap_socket(sock)



import json
import sys

from netlab.client import Client
from paepy.ChannelDefinition import CustomSensorResult

if __name__ == "__main__":
    # interpret first command line parameter as json object
    if len(sys.argv) == 1:
        # for command line debugging, use the default system
        sys_args = '{"params": "default"}'
    else:
        sys_args = sys.argv[1]
    prtg = json.loads(sys_args)
    system = prtg.get('params', "default").lower()

    # PRTG runs sensors as a different user. Specify the Administrator user's netlab config file here
    client = Client(system, config_path='C:/Users/Administrator/.netlab/config.json')

    # query data from netlab api
    response = client.system_usage_disk()
    data_free_space = int(100 - response['data']['used_pct'])
    data_free_bytes = response['data']['total_b'] - response['data']['used_b']
    programs_free_space = int(100 - response['programs']['used_pct'])
    programs_free_bytes = response['programs']['total_b'] - response['programs']['used_b']

    # create sensor result
    result = CustomSensorResult("OK")

    # add primary channel for percentage of data partition free space
    result.add_channel(channel_name="Free Space - data", primary_channel=True, unit='Percent', value=data_free_space,
                       is_limit_mode=True, limit_min_warning=25, limit_min_error=10)

    # additional channels for remaining values
    result.add_channel(channel_name="Free Bytes - data", unit="BytesDisk", value=data_free_bytes)
    result.add_channel(channel_name="Free Space - programs", unit="Percent", value=programs_free_space)
    result.add_channel(channel_name="Free Bytes - programs", unit="BytesDisk", value=programs_free_bytes)

    # print sensor result to stdout
    print(result.get_json_result())



CODES = []


class ResponseFormatError(Exception):
    """The server's response was impossible to decode."""


class InvalidConfig(Exception):
    """Config is not readable or contains errors."""


class InvalidVersion(Exception):
    """The remote API version is not compatible."""


class NetlabError(Exception):
    """Base error from NETLAB+ api."""
    str_error = "E_NETLAB_ERROR"

    def __init__(self, message=None):
        self.message = message

    def __str__(self):
        if self.message is None:
            return "[{}]".format(self.str_error)
        else:
            return '[{}] {}'.format(self.str_error, self.message)


class AuthenticationError(NetlabError):
    """Connection to the NETLAB+ system could not be authenticated."""
    str_error = "E_AUTHENTICATION_FAILED"


class UnmappedError(NetlabError):
    """An Error returned by the server but not handled by this library."""
    def __init__(self, str_error, message):
        self.str_error = str_error
        self.message = message


def _add(cls):
    CODES.append(cls)
    return cls


@_add
class MethodNotFoundError(NetlabError):
    """The requested method could not be found."""
    str_error = "E_METHOD_NOT_FOUND"


@_add
class ParamsNotSupported(NetlabError):
    """Some parameters were incorrect."""
    str_error = "E_PARAMS_NOT_SUPPORTED"


@_add
class MissingParamError(NetlabError):
    """Parameter was missing."""
    str_error = "E_MISSING_PARAM"


@_add
class MissingParamsError(NetlabError):
    """Parameter was missing."""
    str_error = "E_MISSING_PARAMS"


@_add
class InvalidParamsError(NetlabError):
    """Parameters are in some way incorrect."""
    str_error = "E_INVALID_PARAMS"


@_add
class InvalidParamError(NetlabError):
    """Parameter is in some way incorrect."""
    str_error = "E_INVALID_PARAM"


@_add
class InvalidFilterError(NetlabError):
    """Filter is not a valid filter."""
    str_error = "E_FILTER_INVALID"


@_add
class InvalidFilterPropertyError(NetlabError):
    """Filter contains an invalid property or operator."""
    str_error = "E_FILTER_PROPERTY"


@_add
class AccessDeniedError(NetlabError):
    """Caller did not have required privilege."""
    str_error = "E_ACCESS_DENIED"


@_add
class DatesEndBeforeStartError(NetlabError):
    """End date is before start date."""
    str_error = "E_DATES_END_BEFORE_START"


@_add
class ItemDependencyError(NetlabError):
    """The virtual machine is in use by a pod."""
    str_error = "E_ITEM_DEPENDENCY"


@_add
class InvalidTimeError(NetlabError):
    """The time format provided does not match yyy-mm-dd-hh-ss."""
    str_error = "E_INVALID_TIME"


@_add
class TaskNotFoundError(NetlabError):
    """task_id is not found."""
    str_error = "E_TASK_NOT_FOUND"


@_add
class InvalidPropertiesError(NetlabError):
    """The properties requested could not be found."""
    str_error = "E_INVALID_PROPERTIES"


@_add
class LicenseNotActiveError(NetlabError):
    """Software license is not active."""
    str_error = "E_LICENSE_NOT_ACTIVE"



from enum import Enum


class HDRSeverity(Enum):
    """

    """
    NONE = 0
    "Show no warnings."
    FATAL = 1
    "Show FATAL warnings. Level 1."
    ERROR = 2
    "Show FATAL warnings. Level 2."
    WARN = 3
    "Show FATAL warnings. Level 3."
    SUCCESS = 4
    "Show FATAL warnings. Level 4."
    INFO = 5
    "Show FATAL warnings. Level 5."
    TRACE0 = 6
    "Show FATAL warnings. Level 6."
    TRACE1 = 7
    "Show FATAL warnings. Level 7."
    TRACE2 = 8
    "Show FATAL warnings. Level 8."


class HDRStatus(Enum):
    RUNNING = 0
    OK = 1
    FAILED = 2
    PARTIAL = 3
    CANCELLED = 4
    UNKNOWN = 9


class TimeFormat(Enum):
    ""
    HOUR24 = "24"
    "24 Hour"
    HOUR12 = "12"
    "12 Hour (am/pm)"


class DateFormat(Enum):
    ""
    ISO = 'ISO'
    "YYYY-MM-DD"
    DMY_SLASH = 'DMY_SLASH'
    "DD/MM/YYYY"
    DMY_HYPHEN = 'DMY_HYPHEN'
    "DD-MM-YYYY"
    DMY_DOT = 'DMY_DOT'
    "DD.MM.YYYY"
    DMY_SPACE = 'DMY_SPACE'
    "DD MM YYYY"
    DD_MMM_YYYY = 'DD_MMM_YYYY'
    "DD-Mon-YYYY"
    MDY_SLASH = 'MDY_SLASH'
    "MM/DD/YYYY"
    DAY_MDY = 'DAY_MDY'
    "Day Mon D, YYYY"
    CHINESE = 'CHINESE'
    "YYYY M D"
    JAPANESE = 'JAPANESE'
    "YYYY MM DD"



import re

from ..utils import hdr_result, version_gte
from .. import enums

CLONE_SPEC_ATTRS = {
    'pc_type',
    'source_snapshot',
    'clone_snapshot',
    'clone_type',
    'clone_role',
    'clone_name',
    'clone_storage_alloc',
    'clone_datastore',
    'clone_vh_id',
}


class PodApiMixin:
    pod_default_props = ['pod_id', 'pod_name', 'pt_id', 'pod_current_state']
    pod_props = ['pod_id', 'pt_id', 'def_im_name', 'def_topology_image', 'def_vlan_map', 'devices', 'device_count',
                 'pod_acl_enabled', 'pod_admin_state', 'pod_adv_settings', 'pod_auto_net_enabled',
                 'pod_auto_net_host_setup', 'pod_auto_net_host_teardown', 'pod_cat', 'pod_csw_base_port',
                 'pod_csw_base_vlan', 'pod_csw_id', 'pod_current_state', 'pod_desc', 'pod_id', 'pod_managed',
                 'pod_name', 'pod_origin', 'pod_res_id', 'pod_task_type', 'pod_task_id', 'pod_task_info', 'pod_uuid',
                 'pt_actions', 'pt_adv_settings', 'pt_apc_port_count', 'pt_apdid', 'pt_as_port_count', 'pt_author',
                 'pt_build', 'pt_copyright', 'pt_csw_port_count', 'pt_desc', 'pt_gpdid', 'pt_id', 'pt_name', 'pt_notes',
                 'pt_org', 'pt_origin', 'pt_pod_max', 'pt_removable', 'pt_tabs', 'pt_url', 'pt_vlan_pool', 'remote_pc',
                 'remote_pc_count', 'sched_image']

    pod_types_default_props = ['pt_id', 'pt_name', 'pt_build']
    pod_types_props = ['pt_id', 'pt_name', 'pt_build', 'pt_gpdid', 'pt_apdid', 'pt_desc', 'pt_index', 'pt_type',
                       'sched_image', 'def_topology_image', 'def_im_name', 'def_vlan_map', 'pt_tabs', 'pt_actions',
                       'pt_origin', 'pt_vlan_pool', 'pt_pod_max', 'pt_notes', 'pt_csw_port_count', 'pt_as_port_count',
                       'pt_apc_port_count', 'pt_build', 'pt_author', 'pt_org', 'pt_copyright', 'pt_url', 'pt_removable',
                       'pt_ae_pod', 'pt_adv_settings', 'remote_pc_count', 'device_count']

    def __pc_clone_specs(self, *, user_pc_clone_specs, remote_pc_list, clone_pod_name):
        # Ensure that user_pc_clone_specs is a list of dicts
        if isinstance(user_pc_clone_specs, dict):
            user_pc_clone_specs = [user_pc_clone_specs] * len(remote_pc_list)
        elif user_pc_clone_specs is None:
            user_pc_clone_specs = [{}] * len(remote_pc_list)

        if len(user_pc_clone_specs) != len(remote_pc_list):
            raise ValueError("`user_pc_clone_specs` and `remote_pc_list` must be the same length.")

        # Ensure that the remote_pc_list is in order
        remote_pc_list = sorted(remote_pc_list, key=lambda e: e['pl_index'])

        # Result of this function
        new_vm_out = []

        for pl_index, vms in enumerate(zip(user_pc_clone_specs, remote_pc_list), 1):
            user_vm, existing_vm = vms
            bad_clone_specs = set(user_vm.keys()) - CLONE_SPEC_ATTRS
            if bad_clone_specs:
                raise ValueError("These keys are not allowed: {}".format(bad_clone_specs))
            new_vm_out.append({
                'pl_index': pl_index,
                'pc_type': user_vm.get('pc_type', existing_vm['pc_type']),
                'source_vm_id': existing_vm['vm_id'],
                'source_snapshot': user_vm.get('source_snapshot', existing_vm['vm_snapshot']),
                'clone_snapshot': user_vm.get('clone_snapshot', existing_vm['vm_snapshot']),
                'clone_type': user_vm.get('clone_type', 'LINKED'),
                'clone_role': user_vm.get('clone_role', 'NORMAL'),
                'clone_name': user_vm.get('clone_name', clone_pod_name + '_' + existing_vm['pc_label']),
                'clone_storage_alloc': user_vm.get('clone_storage_alloc', 'ONDEMAND'),
                'clone_datastore': user_vm.get(
                    'clone_datastore',
                    re.match('^\[(.*)\].*', existing_vm['vm_path']).group(1)),
                'clone_vh_id': user_vm.get('clone_vh_id', existing_vm['vh_id']),
            })

        return new_vm_out

    def pod_add(self, *, pod_id, pt_id, devices=None, pod_acl_enabled=False, pod_adv_settings=None,
                pod_auto_net_enabled=True, pod_auto_net_host_setup=True, pod_auto_net_host_teardown=True,
                pod_cat=enums.PodCategory.PERSISTENT_VM, pod_desc=None,
                pod_name=None, **kwargs):
        """
        This method allows you to add a pod to the NETLAB+ system. Pods created will be OFFLINE after added.

        :param int pod_id: Unique pod identifier.
        :param str pt_id: Pod type identifier.
        :param list devices: Array of devices associated with this pod. This is for real equipment pods only.
        :param bool pod_acl_enabled: Pod access control list enabled.
        :param str pod_adv_settings: Advanced settings for pod configuration.
        :param bool pod_auto_net_enabled: Auto-networking enabled.
        :param bool pod_auto_net_host_setup: Auto-networking host setup enabled.
        :param bool pod_auto_net_host_teardown: Auto-networking host teardown enabled.
        :param PodCategory pod_cat: Pod category type.
        :param str pod_desc: Pod description.
        :param str pod_name: Pod name.

        :return: The addition succeeded.
        :rtype: 'OK'
        """

        method = "pod.add"
        param = {
            'pod_id': pod_id,
            'pt_id': pt_id,
        }

        if devices is not None:
            param['devices'] = devices
        if pod_acl_enabled is not None:
            param['pod_acl_enabled'] = pod_acl_enabled
        if pod_adv_settings is not None:
            param['pod_adv_settings'] = pod_adv_settings
        param['pod_auto_net_enabled'] = pod_auto_net_enabled
        param['pod_auto_net_host_setup'] = pod_auto_net_host_setup
        param['pod_auto_net_host_teardown'] = pod_auto_net_host_teardown
        param['pod_cat'] = pod_cat
        if pod_desc is not None:
            param['pod_desc'] = pod_desc
        if pod_name is not None:
            param['pod_name'] = pod_name

        param.update(kwargs)
        return self.call(method, **param)

    def pod_get(self, *, pod_id, properties="default", **kwargs):
        """

        :param int pod_id: Unique pod identifier.
        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.
        :type properties: list[str] or 'all' or 'default'

        :return: Returns a dictionary object containing the following properties:
        :rtype: dict

        **Properties**

        pod_id
            Unique pod identifier.
        pt_id
            Pod typed identifier.
        def_im_name
            Default image name.
        def_topology_image
            Default topology image URL.
        def_vlan_map
            Default VLAN mapping.
        devices
            List of devices in pod type.
        device_count
            Count of devices in pod type.
        pod_acl_enabled
            Pod access control list enabled.
        pod_admin_state
            Administrative pod states.
        pod_adv_settings
            A list of all advanced settings.
        pod_auto_net_enabled
            Automatic VM networking enabled.
        pod_auto_net_host_setup
            Automatic Host networking setup enabled.
        pod_auto_net_host_teardown
            Automatic Host networking teardown enabled.
        pod_cat
            Pod category type. :py:class:`netlab.enums.PodCategory`
        pod_csw_base_port
            Pod control switch base port.
        pod_csw_base_vlan
            Pod control switch base VLAN.
        pod_csw_id
            Pod control switch id.
        pod_current_state
            Pod current state status.
        pod_desc
            Pod description.
        pod_id
            Unique pod identifier.
        pod_managed
            Pod is managed.
        pod_name
            Pod name.
        pod_origin
            Pod origin.
        pod_res_id
            Current pod reservation identifier.
        pod_task_type
            Current pod task type.
        pod_task_id
            Current pod task identifier.
        pod_task_info
            Current pod task information.
        pod_uuid
            Globally unique pod identifier.
        pt_actions
            Pod type actions enabled.
        pt_adv_settings
            Long string of advanced settings.
        pt_apc_port_count
            Switched outlet port count requirement.
        pt_apdid
            Author identifier for pod design.
        pt_as_port_count
            Access server port count requirement.
        pt_author
            Pod design author.
        pt_build
            Pod build number.
        pt_copyright
            Pod design copyright.
        pt_csw_port_count
            Control switch port count requirement.
        pt_desc
            Pod type description.
        pt_gpdid
            Global unique identifier.
        pt_id
            Pod type identifier.
        pt_name
            Pod type name.
        pt_notes
            Pod type notes.
        pt_org
            Pod design organization.
        pt_origin
            Pod type origin.
        pt_pod_max
            Max pod count of this type.
        pt_removable
            Pod type is removable.
        pt_tabs
            Pod type tabs enabled.
        pt_url
            Pod design support URL.
        pt_vlan_pool
            Pod type vlan pool.
        remote_pc
            List of remote PCs in pod type with their expanded properties. See Remote PC table below.
        remote_pc_count
            Count of remote PCs in pod type.
        reservations
            The number of reservations scheduled or in progress for this pod.
        sched_image
            Scheduler image URL.
        """

        method = "pod.get"

        if properties == 'all':
            properties = self.pod_props
        if properties == 'default':
            properties = self.pod_default_props

        res_param = {
            'pod_id': pod_id,
            'properties': properties,
        }
        res_param.update(kwargs)
        return self.call(method, **res_param)

    def pod_list(self, *, pod_cat=None, **kwargs):
        """
        List pods on the system.

        :param PodCategory pod_cat: Pod category type.
        :return: Returns a list of dictionary objects containing the following properties:
        :rtype: list[dict]

        **Properties**

        def_topology_image
            Default topology image URL.
        pod_acl_enabled
            Pod access control lists enabled for this pod.
        pod_admin_state
            Pod administrative state. :py:class:`netlab.enums.PodAdminState`
        pod_cat
            Pod category. :py:class:`netlab.enums.PodCategory`
        pod_current_state
            Pod current state. :py:class:`netlab.enums.PodCurrentState`
        pod_desc
            Pod description.
        pod_dyn_vlan
            Pod is using dynamic vlans.
        pod_id
            Unique pod identifier.
        pod_managed
            Pod is managed (currently always True).
        pod_name
            Pod name.
        pod_uuid
            Pod name.
        pod_res_id
            Current pod reservation identifier.
        pt_apdid
            Author identifier for pod design.
        pt_desc
            Pod type description.
        pt_gpdid
            Global unique identifier.
        pt_id
            Pod type identifier.
        pt_name
            Pod type name.
        sched_image
            Scheduler image URL.
        """
        method = "pod.list"

        res_param = {}
        if pod_cat is not None:
            res_param['pod_cat'] = pod_cat

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def pod_list_used_ids(self, **kwargs):
        """
        List the used pod_id's on the system.

        :return: Returns a list of pod_ids
        :rtype: list[int]
        """

        method = "pod.list.used_ids"
        return list(map(int, self.call(method, **kwargs)))

    def pod_pc_get(self, *, pod_id=None, pl_index=None, pc_id=None, **kwargs):
        """
        Get information regarding a pod pc.

        :param int pod_id: Pod ID
        :param int pl_index: PC Pod index
        :param int pc_id: unique pc identifier

        :return: On success, returns the following properties:
        :rtype: dict

        **Properties**

        pc_icon
            Icon identifier. :py:class:`netlab.enums.PCIcon`
        pc_id
            Unique pc identifier.
        pc_label
            PC label.
        pc_online
            PC online.
        pc_origin
            PC origin.
        pc_os_id
            PC operating system identifier.
        pc_os_name
            NETLAB+ operating system name.
        pc_pod_id
            pod_id pc is associated to.
        pc_pod_index
            Index of pods.
        pc_revert_to_scrub
            Revert pc during scrub.
        pc_type
            PC type. :py:class:`netlab.enums.PCType`
        pc_vnc_ip_addr
            IP Address of vnc (ESXi host) server.
        pc_vnc_port
            Port number VNC is running on.
        pc_vnc_use_copy_rect
            Always Allow VNC to use copy rect encoding.
        pl_icon
            Pod design PC icon identifier. :py:class:`netlab.enums.PLIcon`
        pl_index
            PC pod index.
        pl_label
            Remote pc label.
        pt_id
            Pod type identifier.
        vdc_id
            Agent hostname/IP address.
        vdc_name
            VM Datacenter name.
        vh_bios_date
            Timestamp of BIOS creation date.
        vh_bios_version
            BIOS version installed.
        vh_com_path
            Virtual host communication path. :py:class:`netlab.enums.VirtualHostComPath`
        vh_cpu_cores
            Total count of CPU cores on virtual host.
        vh_cpu_mhz
            Speed in MHz of CPU.
        vh_cpu_model
            CPU model identification.
        vh_cpu_threads
            Number of threads CPUs can handle.
        vh_cpu_threads
            Number of threads CPUs can handle.
        vh_date_modified
            Timestamp of virtual host information has been modified.
        vh_date_tested
            Timestamp of virtual host test communication.
        vh_id
            Virtual host identifier.
        vh_inside_ipv4_addr
            Inside interface ipv4 address.
        vh_inside_vswitch_0
            Virtual host inside vSwitch.
        vh_last_test_status
            Status of last virtual host test communication.
        vh_memory_mb
            Total RAM in MB of virtual host.
        vh_name
            Virtual host hostname.
        vh_online
            Virtual host is online.
        vh_os_build
            Operating system build number.
        vh_os_description
            Operating system description.
        vh_os_name
            Operating system name.
        vh_os_type
            Operating system type.
        vh_os_vendor
            Operating system vendor.
        vh_os_version
            Operating system version.
        vh_outside_ipv4_addr
            Outside interface ipv4 address.
        vh_pra_enabled
            Proactive resource awareness is enabled.
        vh_pra_max_cpu
            Max number of CPUs for PRA.
        vh_pra_max_mem_mb
            Max amount of RAM in MB for PRA.
        vh_pra_max_vm
            Max amount of virtual machines for PRA.
        vh_sys_model
            System model.
        vh_sys_service_tag
            System service tag.
        vh_sys_vendor
            System vendor.
        vh_uuid
            Virtual host unique global identifier.
        vhg_id
            Virtual host group identifier.
        vm_alloc_cpu_n
            Total amount of allocated CPUs.
        vm_alloc_mem_mb
            Total amount of allocated RAM.
        vm_auto_display
            Virtual machine auto display port configuration.
        vm_auto_network
            Virtual machine auto networking.
        vm_auto_settings
            Virtual machine auto settings.
        vm_child_count
            Total count of children VMs.
        vm_comments
            Virtual machine comments from VMX.
        vm_date_added
            Timestamp of virtual machine added to NETLAB+ inventory.
        vm_id
            Virtual machine identifier.
        vm_name
            Virtual machine name.
        vm_netlab_os_id
            NETLAB+ match to vm_vendor_os_id.
        vm_parent_id
            Virtual machine parent identifier, if VM is a link clone.
        vm_parent_snapname
            Virtual machine parent snapshot name, if VM is a link clone.
        vm_path
            Virtual machine path on datastore.
        vm_power_state
            Virtual machine power state.
        vm_role
            Virtual machine role in NETLAB+ inventory. :py:class:`netlab.enums.VirtualMachineRole`
        vm_runtime_vh_id
            Unique identifier of runtime virtual host.
        vm_sanity_checks
            Sanity checks.
        vm_shutdown_pref
            Shutdown preference.
        vm_snapshot
            Virtual machine snapshot name.
        vm_uuid
            Virtual machine unique identifier.
        vm_vendor_os_id
            Vendor operating system identifier.
        vm_vendor_os_name
            Vendor operating system name.
        """

        method = "pod.pc.get"

        res_param = {}
        if pod_id is not None:
            res_param['pod_id'] = pod_id
        if pl_index is not None:
            res_param['pl_index'] = pl_index
        if pc_id is not None:
            res_param['pc_id'] = pc_id

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def pod_state_change(self, *, pod_id, state, **kwargs):
        """
        Sets pod state to either ONLINE, OFFLINE or RESUME.

        :param int pod_id: Pod ID
        :param PodState state: PC Pod State

        :return: The change succeeded.
        :rtype: 'OK'
        """

        method = "pod.state.change"

        res_param = {
            'pod_id': pod_id,
            'state': state,
        }

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def pod_types_get(self, *, pt_id, properties="default", **kwargs):
        """
        Returns the build, id and name of installed pods on queried system.

        :param str pt_id: Pod Type ID is a alpha/num hash with prefix.
        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.
        :type properties: list[str] or 'default' or 'all'

        :return: Returns all or selected properties from the list initially passed for return.
        :rtype: list

        **Properties**

        pt_id
            Pod type identifier.
        pt_name
            Account email.
        pt_build
            Account Family(Last) name.
        pt_gpdid
            Global unique identifer.
        pt_apdid
            Short unique identifier.
        pt_desc
            Pod type description.
        pt_index
            Pod type index.
        pt_type
            Pod type id
        sched_image
            Location of scheduler image.
        def_topology_image
            Location of the default topology image.
        def_im_name
            Default image map name.
        def_vlan_map
            Default VLAN map.
        pt_tabs
            Pod type tabs.
        pt_actions
            Pod type actions.
        pt_origin
            Pod type origin.
        pt_vlan_pool
            Pod type vlan pool.
        pt_pod_max
            Max amount of pod types allowed on system.
        pt_notes
            Pod type notes.
        pt_csw_port_count
            Number of consecutive ports on control switch required by pod type.
        pt_as_port_count
            Number of access server ports required by pod type.
        pt_apc_port_count
            Number of switched outlet ports required by pod type.
        pt_build
            Pod type build number.
        pt_author
            Pod type author.
        pt_org
            Pod type organization.
        pt_copyright
            Pod type copyright.
        pt_url
            Pod type URL address.
        pt_removable
            Allows pod type to be uninstalled.
        pt_ae_pod
            Pod type is for NETLAB+ AE.
        pt_adv_settings
            Pod type advanced settings.
        hotspots
            Returns array of hotspot information for **pt_id**.
        remote_pc_layout
            Returns array of remote PC layout information for **pt_id**.
        remote_pc_count
            Get a count of the remote PCs per pod type.
        device_layout
            Returns array of device layout information for **pt_id**.
        device_count
            Get count of the devices for pod type.
        pod_cat_values
            Returns array of pod categories.
        cables
            Returns array of cable connections for pod type.
        """

        method = "pod.types.get"

        if properties == 'all':
            properties = self.pod_types_props
        if properties == 'default':
            properties = self.pod_types_default_props

        res_param = {
            'pt_id': pt_id,
            'properties': properties
        }

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def pod_types_list(self, *, properties="default", **kwargs):
        """
        Obtain a list of pod types.

        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.
        :type properties: list[str] or 'default' or 'all'
        :return: On success, returns [rt_id, pt_build, pt_name] by default, or the following properties if requested.
        :rtype: list

        **Properties**

        pt_id
            Pod type identifier.
        pt_name
            Account email.
        pt_build
            Account Family(Last) name.
        pt_gpdid
            Global unique identifer.
        pt_apdid
            Short unique identifier.???
        pt_desc
            Pod type description.
        pt_index
            Pod type index.
        pt_type
            ???
        sched_image
            Location of scheduler image.
        def_topology_image
            Location of the default topology image.
        def_im_name
            Default name ???
        def_vlan_map
            Default VLAN map.
        pt_tabs
            Pod type tabs.
        pt_actions
            Pod type actions.
        pt_origin
            Pod type origin.
        pt_vlan_pool
            Pod type vlan pool.
        pt_pod_max
            Max amount of pod types allowed on system.
        pt_notes
            Pod type notes.
        pt_csw_port_count
            Number of consecutive ports on control switch required by pod type.
        pt_as_port_count
            Number of access server ports required by pod type.
        pt_apc_port_count
            Number of switched outlet ports required by pod type.
        pt_build
            Pod type build number.
        pt_author
            Pod type author.
        pt_org
            Pod type organization.
        pt_copyright
            Pod type copyright.
        pt_url
            Pod type URL address.
        pt_removable
            Allows pod type to be uninstalled.
        pt_ae_pod
            Is pod type for NETLAB+ AE???.
        pt_adv_settings
            Pod type advanced settings.
        remote_pc_count
            Get count of the remote PCs per pod type.
        device_count
            Get count of the devices per pod type.

        """
        method = "pod.types.list"

        if properties == 'all':
            properties = self.pod_types_props
        if properties == 'default':
            properties = self.pod_types_default_props

        return self.call(method, properties=properties, **kwargs)

    def pod_remove_task(self, *, pod_id, remove_vms=enums.RemoveVMS.NONE, severity_level=enums.HDRSeverity.WARN,
                        **kwargs):
        """
        :param int pod_id: Pod unique identifier.
        :param RemoveVMS remove_vms: Specifies the method for removing the virtual machines.
        :param HDRSeverity severity_level: Display detailed events for the task only at this severity level or
            higher.

        :return: HDR result
        :rtype: HDRStatus
        """

        method = "pod.remove.task"

        params = {
            'pod_id': pod_id,
            'remove_vms': remove_vms,
            'notify_progress': False,
            'notify_complete': False,
        }

        params.update(kwargs)
        results = self.call(method, **params)

        if version_gte(self._server_version, '17.1.3'):
            return hdr_result(results, severity_level)
        else:
            return results

    def pod_clone_task(self, *, source_pod_id, clone_pod_id, clone_pod_name, pc_clone_specs=None,
                       severity_level=enums.HDRSeverity.WARN, **kwargs):
        """
        Clone a pod.

        :param int source_pod_id: Source pod unique identifier.
        :param int clone_pod_id: Cloned pod unique identifier
        :param str clone_pod_name: Cloned unique pod name.
        :param pc_clone_specs: Array of PC clone specifications. Any unpassed params will be taken from the existing
            pods. If a dict is passed, the params will be used on every pod.
        :type pc_clone_specs: list or dict
        :param HDRSeverity severity_level: Display detailed events for the task only at this severity level or
            higher.
        :return: HDR result
        :rtype: HDRStatus

        **pc_clone_specs**

        The format requires a list of dictionaries.
        One dictionary per each PC in the pod, e.g. `[{pc.clone.spec.1}, {pc.clone.spec.2}]`.
        The following is a list of the keys, the expected value type and a key description.

        **PC Specs: (Obtained via remote_pc lookup - no need to adjust)**

        :pc_type: str : [AVMI, ABSENT]
        :source_snapshot: str : default(vm_snapshot)
        :clone_snapshot: str : default(vm_snapshot)
        :clone_type: str : default(LINKED) [LINKED, FULL]
        :clone_role: str : default(NORMAL) [MASTER, NORMAL, PERSISTENT]
        :clone_name: str : default({clone_pod_name}_{pc_label})
        :clone_storage_alloc: str : default(ONDEMAND) [ONDEMAND, PREALLOCATED]
        :clone_datastore: Datastore in which to clone.
        :clone_vh_id: int : RemotePC vh_id

        **Example:**

        ::

            pc_clone_specs = [{
                    "clone_role":"NORMAL",
                    "clone_type":"LINKED",
                    "source_snapshot":"GOLDEN_MASTER",
                    "pc_type":"AVMI",
                    "clone_vh_id":1,
                    "clone_name":"Test_Linux_Install_pod04_CentOS Server",
                    "clone_storage_alloc":"ONDEMAND",
                    "clone_datastore":"local",
                    "clone_snapshot":"GOLDEN_MASTER"
                }, {
                    "clone_role":"NORMAL",
                    "clone_type":"LINKED",
                    "source_snapshot":"GOLDEN_MASTER",
                    "pc_type":"AVMI",
                    "clone_vh_id":1,
                    "clone_name":"Test_Linux_Install_pod04_Ubuntu Workstation",
                    "clone_storage_alloc":"ONDEMAND",
                    "clone_datastore":"local",
                    "clone_snapshot":"GOLDEN_MASTER"
                }]

        """

        method = "pod.clone.task"
        remote_pc = self.pod_get(pod_id=source_pod_id, properties=['remote_pc'])

        # Transferring to internal pc_clone_specs builder.
        pc_clone_specs = self.__pc_clone_specs(clone_pod_name=clone_pod_name,
                                               user_pc_clone_specs=pc_clone_specs,
                                               remote_pc_list=remote_pc['remote_pc'])

        results = self.call(method, source_pod_id=source_pod_id,
                            clone_pod_id=clone_pod_id, clone_pod_name=clone_pod_name,
                            pc_clone_specs=pc_clone_specs, **kwargs)

        if version_gte(self._server_version, '17.1.4'):
            return hdr_result(results, severity_level)
        else:
            return results

    def pod_acl_add(self, *, com_id, pod_id, acc_id=None, cls_id=None, team=None, **kwargs):
        """
        This method allows you to add access controls to a pod on the NETLAB+ system.

        :param int com_id: Community identifier.
        :param int pod_id: Pod identifier.
        :param int acc_id: Account identifier.
        :param int cls_id: Class identifier.
        :param str team: Team assignment.
        :return: Pod access control list UUID.
        :rtype: str
        """

        args = {
            'com_id': com_id,
            'pod_id': pod_id,
            'acc_id': acc_id,
            'cls_id': cls_id,
            'team': team,
        }
        args.update(kwargs)
        return self.call('pod.acl.add', **args)

    def pod_acl_admin_add(self, *, acc_id, pod_id, **kwargs):
        """
        This method allows you add additional pod access control administrators.

        :param int acc_id: Account identifier.
        :param int pod_id: Pod identifier.
        :return: acc_id was authorized for pod_id.
        :rtype: 'OK'
        """

        return self.call('pod.acl.admin.add', acc_id=acc_id, pod_id=pod_id, **kwargs)

    def pod_acl_admin_auth(self, *, acc_id, pod_id, **kwargs):
        """
        This method returns true if a specified instructor account is authorized to manage POD
        ACLs on a specified pod.

        :param int acc_id: Account identifier.
        :param int pod_id: Pod identifier.

        :return: acc_id is authorized for pod_id.
        :rtype: bool
        """

        return bool(int(self.call('pod.acl.admin.auth', acc_id=acc_id, pod_id=pod_id, **kwargs)))

    def pod_acl_admin_list(self, *, pod_id, **kwargs):
        """
        This method allows you to retrieve a list of pod access controls per pod.

        :param int pod_id: Pod identifier.
        :return: Pod access control list.
        :rtype: list
        """

        return self.call('pod.acl.admin.list', pod_id=pod_id, **kwargs)

    def pod_acl_admin_pods(self, *,  acc_id, **kwargs):
        """
        This method allows you to retrieve a list of pods a user has access to.

        :param int acc_id: Account identifier.
        :return: list
        """

        return self.call('pod.acl.admin.pods', acc_id=acc_id, **kwargs)

    def pod_acl_admin_remove(self, *, acc_id, pod_id, **kwargs):
        """
        This method allows you to remove authorization for an account.

        :param int acc_id: Account identifier.
        :param int pod_id: Pod identifier.
        :return: Pod authorization was successfully removed.
        :rtype: 'OK'
        """

        return self.call('pod.acl.admin.remove', acc_id=acc_id, pod_id=pod_id, **kwargs)

    def pod_acl_challenge(self, *, com_id, pod_id, acc_id=None, cls_id=None, team=None, **kwargs):
        """
        This method allows you to test against a pod access control list.

        :param int com_id: Community identifier.
        :param int pod_id: Pod identifier.
        :param int acc_id: Account identifier.
        :param int cls_id: Class identifier.
        :param str team: Team assignment.
        :return: If matching rule found, pod access control list UUID.
        :rtype: uuid
        """

        return self.call('pod.acl.challenge', com_id=com_id, pod_id=pod_id, acc_id=acc_id, cls_id=cls_id,
                         team=team, **kwargs)

    def pod_acl_get(self, *, pacl_uuid, **kwargs):
        """
        This method allows you to retrieve a pod access control list.

        :param int pacl_uuid: Pod access control list UUID.
        :return: Pod access control list.
        :rtype: list
        """

        return self.call('pod.acl.get', pacl_uuid=pacl_uuid, **kwargs)

    def pod_acl_list(self, *, pod_id, **kwargs):
        """
        This method allows you to retrieve a list of pod access controls per pod.

        :param int pod_id: Pod identifier.
        :return: Pod access control list.
        :rtype: list
        """

        return self.call('pod.acl.list', pod_id=pod_id, **kwargs)

    def pod_acl_remove(self, *, pacl_uuid, **kwargs):
        """
        This method allows you to remove a pod access control list.

        :param int pacl_uuid: Pod access control list UUID.
        :return: Pod was successfully removed.
        :rtype: 'OK'
        """

        return self.call('pod.acl.remove', pacl_uuid=pacl_uuid, **kwargs)



from datetime import datetime, timedelta

from .. import enums


class ReservationApiMixin:
    @staticmethod
    def reservation_timeslot_adjuster(*, user_time, round_low=False):
        """
        Round a datetime to the nearest half hour, which is required for some times in netlab.

        :param datetime.datetime user_time: Must be a datetime object. Adjusted for fulfilling time slot
            if not already time slot acceptable.
        :param bool round_low: True: rounds time in minutes down to the low timeslot. False: rounds time up to next
            timeslot.

        :return: Datetime rounded to half hour.
        :rtype: datetime.datetime
        """

        if round_low:
            adjusted_minutes = 30 * (user_time.minute // 30)  # floor division used to round minutes down
            adjusted_time = user_time.replace(minute=adjusted_minutes)
        else:
            adjusted_time = user_time + (datetime.min - user_time) % timedelta(minutes=30)
        return adjusted_time.replace(second=0, microsecond=0)

    def reservation_make(self, *, type, pod_id, cls_id=None, end_time, start_time=None, acc_id=None,
                         reserver_id=None, tz_id=None, tz_olson=None, pt_id=None, ex_id=None,
                         init_config=enums.InitConfig.NONE, **kwargs):
        """
        :param ReservationType type: Type of reservation.
        :param int pod_id: Pod identifier.
        :param int cls_id: Class identifier. Not required where **type** is 'INSTRUCTOR'.
        :param datetime.datetime end_time: Local time end of reservation. This must be on a half hour slot boundary.
        :param datetime.datetime start_time: Local time start of reservation. If not specified it will use the current
            datetime.
        :param int tz_id: Timezone identifier.
        :param str tz_olson: IANA timezone name. If not specified, the timezone of the user identified by acc_id will be
            used.
        :param int pt_id: Pod type identifier.
        :param int ex_id: Exercise identifier.
        :param int acc_id: Account identifier for the reservation. Defaults to **reserver_id**.
        :param InitConfig init_config: Initial configuration.
        :param int reserver_id: Reserver identifier. Defaults to the request user. An admin can use a **reserver_id** to
            make reservations for others.

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        res_id
            Reservation identifier.
        res_start_utc
            Reservation start in UTC.
        res_end_utc
            Reservation end in UTC.
        res_minutes
            Reservation time in minutes
        """

        method = "reservation.make"

        # Because for some reason these are diffrent
        if type is enums.ReservationType.INDIVIDUAL:
            type = 'INDIVIDUAL'
        elif type is enums.ReservationType.ILT_CLASS:
            type = 'ILT_CLASS'
        elif type is enums.ReservationType.INSTRUCTOR:
            type = 'INSTRUCTOR'
        elif type is enums.ReservationType.TEAM:
            type = 'TEAM'

        res_param = {
            'type': type,
            'pod_id': pod_id,
            'end_time': end_time,
        }

        if start_time:
            res_param['start_time'] = start_time
        if acc_id:
            res_param['acc_id'] = acc_id
        if reserver_id:
            res_param['reserver_id'] = reserver_id
        if tz_id:
            res_param['tz_id'] = tz_id
        if tz_olson:
            res_param['tz_olson'] = tz_olson
        if pt_id:
            res_param['pt_id'] = pt_id
        if ex_id:
            res_param['ex_id'] = ex_id
        if cls_id:
            res_param['cls_id'] = cls_id
        if init_config:
            res_param['init_config'] = init_config

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def reservation_get(self, *, res_id, **kwargs):
        """
        Get parameters of a reservation.

        :param int res_id: Reservation identifier.
        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        acc_com_id
            Community identifier.
        acc_full_name
            Account display name.
        acc_id
            Account identifier.
        cls_div_id
            Community identifier.
        cls_id
            Class identifier.
        cls_name
            Class name.
        ex_id
            Lab exercise identifier.
        ex_name
            Lab exercise name.
        pod_desc
            Pod description.
        pod_id
            Unique pod identifier.
        pod_name
            Pod name.
        pt_desc
            Pod type description.
        pt_id
            Pod type identifier.
        pt_name
            Account email.
        res_acc_id
            Account ID.
        res_active_time
            Active reservation timestamp.
        res_asm_data
            ???
        res_attend_time
            Time reservation was attended.
        res_cls_id
            Class identifier.
        res_com_id
            Community identifier.
        res_config
            Initial configuration.
        res_done
            Reservation is done.
        res_end
            Timestamp scheduled end of reservation.
        res_ex_id
            Reservation lab exercise identifier.
        res_flags
            ???
        res_id
            Reservation identifier.
        res_init_fail
            Count of failures during initialization.
        res_is_active
            Reservation is active.
        res_minutes
            Count of minutes for reservation.
        res_pod_id
            Reservation pod identifier.
        res_post_time
            Timestamp of actual end of reservation.
        res_preload_time
            Timestamp UTC lab will preload.
        res_pt_id
            Reservation pod type identifier.
        res_remaining_dhms
            Days, hours, min, sec remaining if pod is in ACTIVE_LOAD or ACTIVE_INIT state.
        res_remaining_sec
            Seconds remaining if pod is in ACTIVE_LOAD or ACTIVE_LAB state.
        res_start
            Timestamp to start reservation.
        res_team
            Reservation team setting.
        res_type
            Reservation type. :py:class:`netlab.enums.ReservationType`
        res_uuid
            Reservation unique identifier.
        sched_image
            Location of the scheduler image.
        """

        method = "reservation.get"
        res_param = {'res_id': res_id}

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def reservation_post(self, *, res_id, **kwargs):
        """
        Ends a reservation that has already started.

        :param int res_id: Reservation identifier.

        :return: The change succeeded.
        :rtype: 'OK'
        """

        method = "reservation.post"
        return self.call(method, res_id=res_id, **kwargs)

    def reservation_cancel(self, *, res_id, **kwargs):
        """
        Cancel a reservation in NETLAB+ prior to reservation being started.

        NOTE - if you wish to end a reservation that is already started use reservation_post.

        :param int res_id: Reservation unique identifier.

        :return: The change succeeded.
        :rtype: 'OK'
        """

        method = "reservation.cancel"
        return self.call(method, res_id=res_id, **kwargs)

    def reservation_summary(self, **kwargs):
        """
        This method allows you to retrieve a summary of all reservations.

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        active_reservations
            Count of active reservations.
        completed_reservations
            Count of completed reservations.
        future_reservations
            Count of future reservations.
        pods_in_use
            Count of pods in use.
        total_reservations
            Count of total reservations.

        """

        method = "reservation.summary"
        return self.call(method, **kwargs)

    def reservation_query(self, *, scope=enums.ReservationScope.ALL, min_time=None, max_time=None, com_id=None,
                          cls_id=None, pod_id=None, active=None, **kwargs):
        """
        Return a list of reservations filtered on criteria.

        :param ReservationScope scope: Reservation scope.
        :param bool active: Filter active reservations.
        :param int cls_id: Class identifier.
        :param int com_id: Community identifier.
        :param datetime.datetime max_time: Maximum time range.
        :param datetime.datetime min_time: Minimum time range.
        :param int pod_id: Pod identifier.

        :return: On success, returns an object with the following properties:
        :rtype: list[dict]

        **Properties**

        acc_com_id
            Community identifier.
        acc_full_name
            Account display name.
        acc_id
            Account identifier.
        cls_div_id
            Community identifier.
        cls_id
            Class identifier.
        cls_name
            Class name.
        ex_id
            Lab exercise identifier.
        ex_name
            Lab exercise name.
        pod_desc
            Pod description.
        pod_id
            Unique pod identifier.
        pod_name
            Pod name.
        pt_desc
            Pod type description.
        pt_id
            Pod type identifier.
        pt_name
            Account email.
        res_acc_id
            Account identifier.
        res_active_time
            Active reservation timestamp.
        res_attend_time
            Time reservation was attended.
        res_cls_id
            Class identifier.
        res_com_id
            Community identifier.
        res_config
            Initial configuration.
        res_done
            Reservation is done.
        res_end
            Timestamp scheduled end of reservation.
        res_ex_id
            Reservation lab exercise identifier.
        res_flags
            Reservation hex code for pod state.
        res_id
            Reservation identifier.
        res_init_fail
            Count of failures during initialization.
        res_is_active
            Reservation is active.
        res_minutes
            Count of minutes for reservation.
        res_pod_id
            Reservation pod identifier.
        res_post_time
            Timestamp of actual end of reservation.
        res_pt_id
            Reservation pod type identifier.
        res_start
            Timestamp to start reservation.
        res_team
            Reservation team setting.
        res_type
            Type of reservation. :py:class:`netlab.enums.ReservationType`
        res_uuid
            Reservation unique identifier.
        sched_image
            Location of the scheduler image.
        """

        method = "reservation.query"

        res_param = {}
        if active:
            res_param['active'] = active
        if cls_id:
            res_param['cls_id'] = cls_id
        if com_id:
            res_param['com_id'] = com_id
        if max_time:
            res_param['max_time'] = max_time
        if min_time:
            res_param['min_time'] = min_time
        if pod_id:
            res_param['pod_id'] = pod_id
        if scope:
            res_param['scope'] = scope

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def reservation_time_now(self, *, tz_id=None, tz_olson=None, **kwargs):
        """
        This method returns the current NETLAB time.

        :param int tz_id: Timezone identifier.
        :param str tz_olson:

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        local
            Local date time.
        local_am_pm
            AM/PM indicator.
        local_day
            Local day of the month.
        local_day_of_week
            Local day of the week.
        local_hmap
            Local time and am/pm.
        local_hms
            Local time.
        local_hour_12
            Local 12 Hour time.
        local_hour_24
            Local 24 Hour time.
        local_minute
            Local minute time.
        local_second
            Local second time.
        local_year
            Local year.
        local_ymd
            Local YYYY-MM-DD.
        tz_id
            Timezone identifier.
        tz_is_dst
            Timezone is in Daylight Savings Time.
        tz_name
            Name of the timezone.
        tz_olson
            Timezone olson.
        utc
            Coordinated Universal Time.
        """

        method = "reservation.time.now"
        res_param = {}

        if not (tz_id is None or tz_id is 0):  # tz_id==0 is utc, but this fails if passed to api
            res_param['tz_id'] = tz_id
        if tz_olson is not None:
            res_param['tz_olson'] = tz_olson

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def reservation_time_offset(self, *, time=None, op=None, tz_id_in=None,
                                tz_olson_in=None, tz_id_out=None, tz_olson_out=None,
                                slot_roundup=False, years=None, months=None,
                                weeks=None, days=None, hours=None,  minutes=None,
                                seconds=None, **kwargs):
        """
        This method returns the current time with a timezone offset.

        :param datetime time: Time input.
        :param str op: Operator. "+" rounds up. "-" rounds down.
        :param int tz_id_in: Timezone identifier input.
        :param str tz_olson_in: Timezone olson input.
        :param int tz_id_out: Timezone identifier output.
        :param str tz_olson_out: Timezone olson output.
        :param bool slot_roundup: Round up to nearest slot. i.e. 30 min interval. Defaults to False.
        :param int years: Number of years to increase/decrease.
        :param int months: Number of months to increase/decrease.
        :param int weeks: Number of weeks to increase/decrease.
        :param int days: Number of days to increase/decrease.
        :param int hours: Number of hours to increase/decrease.
        :param int minutes: Number of minutes to increase/decrease.
        :param int seconds: Number of seconds to increase/decrease.

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        days
            int
        hours
            int
        minutes
            int
        months
            int
        time_in_local
            datetime
        time_in_utc
            datetime
        time_out_local
            datetime
        time_out_utc
            datetime
        tz_id_in
            int
        tz_id_out
            int
        tz_olson_in
            str
        tz_olson_out
            str
        """

        method = "reservation.time.offset"
        res_param = {'slot_roundup': slot_roundup}

        if time:
            res_param['time'] = time
        if op:
            res_param['op'] = op
        if not (tz_id_in is None or tz_id_in is 0):  # tz_id==0 is utc, but this fails if passed to api
            res_param['tz_id_in'] = tz_id_in
        if tz_olson_in:
            res_param['tz_olson_in'] = tz_olson_in
        if not (tz_id_out is None or tz_id_out is 0):  # tz_id==0 is utc, but this fails if passed to api
            res_param['tz_id_out'] = tz_id_out
        if tz_olson_out:
            res_param['tz_olson_out'] = tz_olson_out
        if years:
            res_param['years'] = years
        if months:
            res_param['months'] = months
        if weeks:
            res_param['weeks'] = weeks
        if days:
            res_param['days'] = days
        if hours:
            res_param['hours'] = hours
        if minutes:
            res_param['minutes'] = minutes
        if seconds:
            res_param['seconds'] = seconds

        res_param.update(kwargs)
        return self.call(method, **res_param)

    def reservation_time_delta(self, *, end, start=None, tz_id=None, tz_olson=None, **kwargs):
        """
        This method returns time between two timestamps with optional timezone offset.

        :param datetime.datetime end: End time. This must be on a half hour increment.
        :param datetime.datetime start: Start time. Defaults to now if not specified.
        :param int tz_id: Timezone identifier.
        :param str tz_olson: Timezone olson.

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        days
            Number of days to increase/decrease.
        hours
            Number of hours to increase/decrease.
        minutes
            Number of minutes to increase/decrease.
        months
            Number of months to increase/decrease.
        """
        method = "reservation.time.delta"

        res_param = {
            'end': end,
        }
        if start:
            res_param['start'] = start
        if not (tz_id is None or tz_id is 0):  # tz_id==0 is utc, but this fails if passed to api
            res_param['tz_id'] = tz_id
        if tz_olson:
            res_param['tz_olson'] = tz_olson

        res_param.update(kwargs)
        return self.call(method, **res_param)



import json
import sys

from netlab.client import Client
from paepy.ChannelDefinition import CustomSensorResult

if __name__ == "__main__":
    # interpret first command line parameter as json object
    if len(sys.argv) == 1:
        # for command line debugging, use the default system
        sys_args = '{"params": "default"}'
    else:
        sys_args = sys.argv[1]
    prtg = json.loads(sys_args)
    system = prtg.get('params', "default").lower()

    # PRTG runs sensors as a different user. Specify the Administrator user's netlab config file here
    client = Client(system, config_path='C:/Users/Administrator/.netlab/config.json')

    # query data from netlab api
    response = client.system_usage_memory()

    total_b = response['ram']['total_b']
    free_b = response['ram']['free_b']
    free_pct = int((free_b / total_b) * 100)

    # create sensor result
    result = CustomSensorResult("OK")

    # add primary channel for percentage of free memory
    result.add_channel(channel_name="Percent Free Memory", primary_channel=True, unit='Percent', value=free_pct)

    # additional channels for remaining values
    result.add_channel(channel_name="Free Memory", unit="BytesMemory", value=free_b)
    result.add_channel(channel_name="Total Memory", unit="BytesMemory", value=total_b)

    # print sensor result to stdout
    print(result.get_json_result())



from .. import enums


class VmApiMixin:
    def vm_datacenter_get(self, *, vdc_id, **kwargs):
        """
        This method allows you to retrieve a datacenter's information by vdc_id.

        :param int vdc_id: VM Datacenter identifier.

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        vdc_agent_hostname
            Agent hostname/IP address.
        vdc_agent_password
            Agent password.
        vdc_agent_username
            Agent username.
        vdc_date_added
            Timestamp of when VDC was added to NETLAB+.
        vdc_date_tested
            Timestamp of when VDC was last tested.
        vdc_last_test_status
            Status of last VDC test.
        vdc_id
            VM Datacenter identifier.
        vdc_name
            VM Datacenter name.
        vdt_type
            VM Datacenter type.
        vdt_name
            VM Datacenter type name.
        """

        method = "vm.datacenter.get"
        return self.call(method, vdc_id=vdc_id, **kwargs)

    def vm_datacenter_find(self, *, vdc_name, **kwargs):
        """
        This method allows you to find a datacenter's **vdc_id** by the datacenter name.

        :param str vdc_name: VM Datacenter name.

        :return: On success, returns vdc_id.
        :rtype: int
        """

        method = "vm.datacenter.get"
        datacenter = self.call(method, vdc_name=vdc_name, **kwargs)
        return datacenter['vdc_id']

    def vm_datacenter_list(self, **kwargs):
        """
        This method allows you to retrieve all datacenters' information.

        :return: A list of objects with the following properties:
        :rtype: list

        **Properties**

        vdc_agent_hostname
            Agent hostname/IP address.
        vdc_agent_password
            Agent password.
        vdc_agent_username
            Agent username.
        vdc_date_added
            Timestamp of when VDC was added to NETLAB+.
        vdc_date_tested
            Timestamp of when VDC was last tested.
        vdc_last_test_status
            Status of last VDC test.
        vdc_id
            VM Datacenter identifier.
        vdc_name
            VM Datacenter name.
        vdt_type
            VM Datacenter type.
        vdt_name
            VM Datacenter type name.
        """

        method = "vm.datacenter.list"
        return self.call(method, **kwargs)

    def vm_datacenter_test(self, *, vdc_id, **kwargs):
        """
        This method allows you to call a task to test communication to datacenter.

        :param int vdc_id: VM Datacenter identifier.

        :return: On success returns 'PASSED', otherwise returns 'FAILED'.
        :rtype: 'PASSED' or 'FAILED'
        """

        method = "vm.datacenter.test.task"
        return self.call(method, vdc_id=vdc_id, **kwargs)

    def vm_host_get(self, *, vh_id, **kwargs):
        """
        This method allows you to retrieve a virtual host information.

        :param int vh_id: Virtual host identifier.

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        vdc_id
            Agent hostname/IP address.
        vdc_name
            VM Datacenter name.
        vdt_type
            VM Datacenter type.
        vdt_name
            VM Datacenter type name.
        vh_bios_date
            Timestamp of BIOS creation date.
        vh_bios_version
            BIOS version installed.
        vh_com_path
            Virtual host communication path.
        vh_cpu_cores
            Total count of CPU cores on virtual host.
        vh_cpu_count
            Total count of physical CPUs on virtual host.
        vh_cpu_mhz
            Speed in MHz of CPU.
        vh_cpu_model
            CPU model identification.
        vh_cpu_threads
            Number of threads CPUs can handle.
        vh_date_added
            Timestamp of virtual host added to NETLAB+.
        vh_date_modified
            Timestamp of virtual host information has been modified.
        vh_date_tested
            Timestamp of virtual host test communication.
        vh_id
            Virtual host identifier.
        vh_inside_ipv4_addr
            Inside interface ipv4 address.
        vh_inside_vswitch_0
            Virtual host inside vSwitch.
        vh_last_test_status
            Status of last virtual host test communication.
        vh_memory_mb
            Total RAM in MB of virtual host.
        vh_name
            Virtual host hostname.
        vh_online
            Virtual host is online.
        vh_os_build
            Operating system build number.
        vh_os_description
            Operating system description.
        vh_os_name
            Operating system name.
        vh_os_type
            Operating system type.
        vh_os_vendor
            Operating system vender.
        vh_os_version
            Operating system version.
        vh_outside_ipv4_addr
            Outside interface ipv4 address.
        vh_pra_enabled
            Proactive resource awareness is enabled.
        vh_pra_max_cpu
            Max number of CPUs for PRA.
        vh_pra_max_mem_mb
            Max amount of RAM in MB for PRA.
        vh_pra_max_vm
            Max amount of virtual machines for PRA.
        vh_sys_model
            System model.
        vh_sys_service_tag
            System service tag.
        vh_sys_vendor
            System vendor.
        vh_uuid
            Virtual host unique global identifier.
        """

        method = "vm.host.get"
        return self.call(method, vh_id=vh_id, **kwargs)

    def vm_host_list(self, *, vdc_id=0, **kwargs):
        """
        This method allows you to retrieve a list of virtual hosts information.

        :param str vdc_id: VM Datacenter identifier. If not specified, will return from all datacenters.

        :return: Returns a list of objects with the following properties:
        :rtype: dict

        **Properties**

        vdc_id
            Agent hostname/IP address.
        vdc_name
            VM Datacenter name.
        vdt_type
            VM Datacenter type.
        vdt_name
            VM Datacenter type name.
        vh_bios_date
            Timestamp of BIOS creation date.
        vh_bios_version
            BIOS version installed.
        vh_com_path
            Virtual host communication path.
        vh_cpu_cores
            Total count of CPU cores on virtual host.
        vh_cpu_count
            Total count of physical CPUs on virtual host.
        vh_cpu_mhz
            Speed in MHz of CPU.
        vh_cpu_model
            CPU model identification.
        vh_cpu_threads
            Number of threads CPUs can handle.
        vh_date_added
            Timestamp of virtual host added to NETLAB+.
        vh_date_modified
            Timestamp of virtual host information has been modified.
        vh_date_tested
            Timestamp of virtual host test communication.
        vh_id
            Virtual host identifier.
        vh_inside_ipv4_addr
            Inside interface ipv4 address.
        vh_inside_vswitch_0
            Virtual host inside vSwitch.
        vh_last_test_status
            Status of last virtual host test communication.
        vh_memory_mb
            Total RAM in MB of virtual host.
        vh_name
            Virtual host hostname.
        vh_online
            Virtual host is online.
        vh_os_build
            Operating system build number.
        vh_os_description
            Operating system description.
        vh_os_name
            Operating system name.
        vh_os_type
            Operating system type.
        vh_os_vendor
            Operating system vender.
        vh_os_version
            Operating system version.
        vh_outside_ipv4_addr
            Outside interface ipv4 address.
        vh_pra_enabled
            Proactive resource awareness is enabled.
        vh_pra_max_cpu
            Max number of CPUs for PRA.
        vh_pra_max_mem_mb
            Max amount of RAM in MB for PRA.
        vh_pra_max_vm
            Max amount of virtual machines for PRA.
        vh_sys_model
            System model.
        vh_sys_service_tag
            System service tag.
        vh_sys_vendor
            System vendor.
        vh_uuid
            Virtual host unique global identifier.
        """

        method = "vm.host.list"
        return self.call(method, vdc_id=vdc_id, **kwargs)

    def vm_inventory_add(self, *, vdc_id, vm_uuid, vm_name, vm_alloc_mem_mb, vm_alloc_cpu_n, vhg_id=None, vh_id=None,
                         vm_role=enums.VirtualMachineRole.NORMAL, vm_comments=None, vm_vendor_os_id=None, **kwargs):
        """
        Add virtual machines to NETLAB+ inventory.

        :param int vdc_id: VM Datacenter identifier.
        :param int vhg_id: Virtual host group identifier.
        :param int vh_id: Virtual host identifier.
        :param str vm_uuid: Virtual machine unique identifier.
        :param str vm_name: Virtual machine name.
        :param str vm_path: Virtual machine path on datastore.
        :param netlab.enums.VirtualMachineRole vm_role Virtual machine role in NETLAB+ inventory.
        :param int vm_alloc_mem_mb: Total amount of allocated RAM.
        :param int vm_alloc_cpu_n: Total amount of allocated CPUs.
        :param str vm_comments: Virtual machine comments from VMX.
        :param str vm_vendor_os_id: Vendor operating system identifier.
        :param str vm_vendor_os_name: Vendor operating system name.
        :param str vm_netlab_os_id: NETLAB+ match to vm_vendor_os_id.
        :param int vm_parent_id: Virtual machine parent identifier, if VM is a link clone.
        :param str vm_parent_snapname: Virtual machine parent snapname, if VM is a link clone.

        :return: Returns a list of object with the following properties:
        :rtype: list[dict]

        **properties**

        vm_id
            Virtual machine identifier.
        """

        method = 'vm.inventory.add'
        return self.call(method, vdc_id=vdc_id, vm_uuid=vm_uuid, vm_name=vm_name, vm_alloc_mem_mb=vm_alloc_mem_mb,
                         vm_alloc_cpu_n=vm_alloc_cpu_n, vhg_id=vhg_id, vh_id=vh_id, vm_role=vm_role,
                         vm_comments=vm_comments, vm_vendor_os_id=vm_vendor_os_id, **kwargs)

    def vm_inventory_get(self, *, vm_id, **kwargs):
        """
        This method allows you to retrieve one or more virtual machines from NETLAB+ inventory.

        :param int vm_id: Virtual machine identifier.

        :return: Returns an object with the following properties:
        :rtype: dict

        **Properties**

        pc_id
            Remote pc identifier.
        pc_label
            Remote pc label.
        pc_os_id
            NETLAB+ operating system identifier.
        pc_os_name
            NETLAB+ operating system name.
        pc_pod_id
            If attached to a pod, pod identifier.
        pc_vm_id
            Virtual machine identifier.
        pod_name
            Pod name.
        vdc_id
            Virtual datacenter identifier.
        vdc_name
            Virtual datacenter name.
        vh_id
            Virtual host identifier.
        vh_name
            Virtual host name.
        vhg_id
            Virtual host group identifier.
        vhg_name
            Virtual host group name.
        vm_alloc_cpu_n
            Total amount of allocated CPUs.
        vm_alloc_mem_mb
            Total amount of allocated RAM.
        vm_child_count
            Total count of children VMs.
        vm_comments
            Virtual machine comments from VMX.
        vm_date_added
            Timestamp of virtual machine added to NETLAB+ inventory.
        vm_id
            Virtual machine identifier.
        vm_name
            Virtual machine name.
        vm_netlab_os_id
            NETLAB+ match to vm_vendor_os_id.
        vm_parent_id
            Virtual machine parent identifier, if VM is a link clone.
        vm_parent_name
            Virtual machine parent name.
        vm_parent_role
            Parent virtual machine role.
        vm_parent_snapname
            Virtual machine parent snapname, if VM is a link clone.
        vm_path
            Virtual machine path on datastore.
        vm_power_state
            Virtual machine power state.
        vm_role
            Virtual machine role in NETLAB+ inventory.
        vm_snapshot
            Virtual machine snapshot name.
        vm_uuid
            Virtual machine unique identifier.
        vm_vendor_os_id
            Vendor operating system identifier.
        vm_vendor_os_name
            Vendor operating system name.

        """
        method = "vm.inventory.get"
        return self.call(method, vm_id=vm_id, **kwargs)

    def vm_inventory_list(self, *, vdc_id=None, roles=None, attached=None, **kwargs):
        """
        This method allows you to list the virtual machines from NETLAB+ inventory.

        :param int vdc_id: VM Datacenter identifier.
        :param list roles: Filter by specified roles.
        :param int attached: List virtual machines attached to a pod.


        :return: Returns an object with the following properties:
        :rtype: dict

        **Properties**

        pc_id
            Remote pc identifier.
        pc_label
            Remote pc label.
        pc_os_id
            NETLAB+ operating system identifier.
        pc_os_name
            NETLAB+ operating system name.
        pc_pod_id
            If attached to a pod, pod identifier.
        pc_vm_id
            Virtual machine identifier.
        pod_name
            Pod name.
        vdc_id
            Virtual datacenter identifier.
        vdc_name
            Virtual datacenter name.
        vh_id
            Virtual host identifier.
        vh_name
            Virtual host name.
        vhg_id
            Virtual host group identifier.
        vhg_name
            Virtual host group name.
        vm_alloc_cpu_n
            Total amount of allocated CPUs.
        vm_alloc_mem_mb
            Total amount of allocated RAM.
        vm_child_count
            Total count of children VMs.
        vm_comments
            Virtual machine comments from VMX.
        vm_date_added
            Timestamp of virtual machine added to NETLAB+ inventory.
        vm_id
            Virtual machine identifier.
        vm_name
            Virtual machine name.
        vm_netlab_os_id
            NETLAB+ match to vm_vendor_os_id.
        vm_parent_id
            Virtual machine parent identifier, if VM is a link clone.
        vm_parent_name
            Virtual machine parent name.
        vm_parent_role
            Parent virtual machine role.
        vm_parent_snapname
            Virtual machine parent snapname, if VM is a link clone.
        vm_path
            Virtual machine path on datastore.
        vm_power_state
            Virtual machine power state.
        vm_role
            Virtual machine role in NETLAB+ inventory.
        vm_snapshot
            Virtual machine snapshot name.
        vm_uuid
            Virtual machine unique identifier.
        vm_vendor_os_id
            Vendor operating system identifier.
        vm_vendor_os_name
            Vendor operating system name.
        """
        method = "vm.inventory.get"
        params = {}

        if vdc_id:
            params['vdc_id'] = vdc_id
        if roles:
            params['roles'] = roles
        if attached is not None:
            params['attached'] = attached

        params.update(kwargs)
        return self.call(method, **params)

    def vm_snapshot_add(self, *, vm_id=None, pc_id=None, snapshot_name, dump_memory=False, sync=False,
                        description=None, **kwargs):
        """
        Method to call a task to add snapshot to virtual machine.

        :param int vm_id: Virtual datacenter identifier.
        :param int pc_id: Remote pc identifier.
        :param str snapshot_name: Name of snapshot to take.
        :param bool dump_memory: Dump memory files.
        :param bool sync: Sync files.
        :param str description: Snapshot description.

        :return: The snapshot completed successfully.
        :rtype: 'OK'
        """

        method = "vm.snapshot.add.task"
        return self.call(method, vm_id=vm_id, pc_id=pc_id, snapshot_name=snapshot_name, dump_memory=dump_memory,
                         sync=sync, description=description, **kwargs)

    def vm_snapshot_edit(self, *, vm_id=None, pc_id=None, snapshot_id=None, new_snapshot_name=None,
                         new_snapshot_description=None, **kwargs):
        """
        Method to call a task to edit a snapshot.

        :param int vm_id: Virtual datacenter identifier.
        :param int pc_id: Remote pc identifier.
        :param int snapshot_id: Snapshot identifier.
        :param int new_snapshot_name: New name of snapshot to take.
        :param str new_snapshot_description: New description of snapshot.

        :return: The edit completed successfully.
        :rtype: 'OK'
        """
        method = "vm.snapshot.edit.task"
        return self.call(method, vm_id=vm_id, pc_id=pc_id, snapshot_id=snapshot_id,
                         new_snapshot_name=new_snapshot_name, new_snapshot_description=new_snapshot_description,
                         **kwargs)

    def vm_snapshot_get_list(self, *, vm_id=None, pc_id=None, **kwargs):
        """
        Method to call a task to retrieve a list of snapshots.

        :param int vm_id: Virtual datacenter identifier.
        :param int pc_id: Remote pc identifier.

        :return: Returns a dict with the following properties:
        :rtype: dict

        **Properties**
        current_snapshot
            The current snapshot. See *Snapshot properties* for more info.
        snapshot_list
            A list of all snapshots. See *Snapshot properties* for more info.

        **Snapshot properties**
        power_state
            The state of the snapshot. :py:class:`PowerState`
        name
            The snapshot name.
        description
            The snapshot description.
        parent_id
            The id of the snapshot parent.
        id
            The snapshot id.
        mo_ref
            The unique identifier of the managed object in vSphere.
        """
        method = "vm.snapshot.get.list.task"
        return self.call(method, vm_id=vm_id, pc_id=pc_id, **kwargs)

    def vm_snapshot_get_tree(self, *, vm_id=None, pc_id=None, **kwargs):
        """
        Method to call a task to retrieve snapshot tree.

        :param int vm_id: Virtual datacenter identifier.
        :param int pc_id: Remote pc identifier.

        :return: Returns a dict with the following properties:
        :rtype: dict

        **Properties**
        current_snapshot
            The current snapshot. See *Snapshot properties* for more info.
        snapshot_list
            A list of all snapshots. See *Snapshot properties* for more info.

        **Snapshot properties**
        power_state
            The state of the snapshot. :py:class:`PowerState`
        name
            The snapshot name.
        description
            The snapshot description.
        id
            The snapshot id.
        mo_ref
            The unique identifier of the managed object in vSphere.
        """
        method = "vm.snapshot.get.tree.task"
        return self.call(method, vm_id=vm_id, pc_id=pc_id, **kwargs)

    def vm_snapshot_remove(self, *, vm_id=None, pc_id=None, snapshot_name=None, snapshot_id=None, remove_children=None,
                           remove_all_snapshots=None, **kwargs):
        """
        Method to call a task to remove snapshot from virtual machine.

        :param int vm_id: Virtual datacenter identifier.
        :param int pc_id: Remote pc identifier.
        :param str snapshot_name: Name of snapshot to take.
        :param int snapshot_id: Snapshot identifier.
        :param bool remove_children: Remove children snapshots.
        :param bool remove_all_snapshots: Remove all snapshots of virtual machine.

        :return: The removal completed successfully.
        :rtype: 'OK'
        """
        method = "vm.snapshot.remove.task"
        return self.call(
            method, vm_id=vm_id, pc_id=pc_id, snapshot_name=snapshot_name, snapshot_id=snapshot_id,
            remove_children=remove_children, remove_all_snapshots=remove_all_snapshots, **kwargs)

    def vm_snapshot_revert(self, *, vm_id=None, pc_id=None, snapshot_name=None, snapshot_id=None, **kwargs):
        """
        Method to call a task to revert to snapshot from virtual machine.

        :param int vm_id: Virtual datacenter identifier.
        :param int pc_id: Remote pc identifier.
        :param str snapshot_name: Name of snapshot to take.
        :param int snapshot_id: Snapshot identifier.

        :return: The revert completed successfully.
        :rtype: 'OK'
        """
        method = "vm.snapshot.revert.task"
        return self.call(
            method, vm_id=vm_id, pc_id=pc_id, snapshot_name=snapshot_name, snapshot_id=snapshot_id, **kwargs)



from enum import Enum


class AccountPrivileges(Enum):
    ""
    COMWIDE = 'COMWIDE'
    "Community Administrator."
    SYSWIDE = 'SYSWIDE'
    "Systemwide Administrator."
    POD_DESIGNER = 'POD_DESIGNER'
    "Allows use of Pod Designer."
    LAB_DESIGNER = 'LAB_DESIGNER'
    "Allows use of Lab Designer."


class AccountType(Enum):
    """
    Account type.
    """
    STUDENT = 'S'
    "Student"
    INSTRUCTOR = 'I'
    "Instructor"
    ADMIN = 'Z'
    "Admin"
